#! /usr/bin/env python
# This file is licensed under the GNU Lesser General Public License
# (LGPL).  See LICENSE.txt for details.

import os, time, math
from math import sin, cos, pi
from copy import copy
import cmath
import time
import itertools
tau = pi * 2

from point import Point
from PointRotate import PointRotate3D

try:
    #import Fred
    import Gnuplot, Gnuplot.PlotItems, Gnuplot.funcutils
    gp = Gnuplot.Gnuplot(debug=1)
except ImportError:
    gp = None


def wait(str=None, prompt='Press return to show results...\n'):
    if str is not None:
        print str
    raw_input(prompt)


def plot(_d, title=None):
    """
        Exercise the Gnuplot module.
    """

    if gp is None:
        return

    d = []
    for ptp in _d:
        np = []
        for p in ptp:
            np.append(p.lst())
        d.append(np)

    gp.clear()
    gp("set zrange [-1:1]")
    gp("set xrange [-5:5]")
    gp("set yrange [-5:5]")
    gp('set style line 1 linecolor rgb "blue"')
    gp('set style line 2 linecolor rgb "blue"')
    gp('set style data lines')

    print '############### test splot ##################################'
    gp.splot(Gnuplot.Data(d, with_='linesp', inline=1))

    if title:
        gp.hardcopy("%s.ps" % title, mode="landscape",
                            color=1, enhanced=1)
        os.system("ps2pdf %s.ps %s.pdf" % (title, title))

    time.sleep(3)


class Operation:
    def __init__(self):
        pass
    def __repr__(self):
        return self.__class__.__name__

zaxis = [Point(0, 0, 0), Point(0, 0, 1.0)]
yaxis = [Point(0, 0, 0), Point(0, 1.0, 0)]
xaxis = [Point(0, 0, 0), Point(1.0, 0, 0)]


class Rotate:
    def __init__(self, axis, ang):
        self.axis = axis
        self.ang = ang

    def fmt(self, ang):
        return "%s<%f>" % (self.__class__.__name__, math.degrees(ang))

    def __repr__(self):
        return self.fmt(self.ang)

    def action_pt(self, pt):
        return [ \
                    PointRotate3D(self.axis[0], self.axis[1], pt[0], self.ang),
                    PointRotate3D(self.axis[0], self.axis[1], pt[1], self.ang)
                ]

    def action_pts(self, pt):
        return self.action_pt(pt)

    def action(self, pts):
        newpts = []
        for pt in pts:
            newpts.append(self.action_pts(pt))
        return newpts


class RotateZ(Rotate):
    def __init__(self, ang):
        Rotate.__init__(self, zaxis, ang)

class RotateY(Rotate):
    def __init__(self, ang):
        Rotate.__init__(self, yaxis, ang)

class RotateX(Rotate):
    def __init__(self, ang):
        Rotate.__init__(self, xaxis, ang)

class PointReverse:

    def __init__(self, yes):
        self.yes = yes

    def action(self, pts):
        if not self.yes:
            return pts

        newpts = []
        for pt in pts:
            pt = copy(pt)
            newpts.append(pt)
        return newpts

    def fmt(self, arg):
        return "%s<%s>" % (self.__class__.__name__, str(arg))

class PointMirrorZ:

    def __init__(self, yes):
        self.yes = yes

    def action(self, pts):
        if not self.yes:
            return pts

        newpts = []
        for pt in pts:
            newpt = []
            for p in pt:
                newpt.append(Point(p.x,p.y,-p.z))
            newpts.append(newpt)
        return newpts

    def fmt(self, arg):
        return "%s<%s>" % (self.__class__.__name__, str(arg))

class PointMirrorX:

    def __init__(self, yes):
        self.yes = yes

    def action(self, pts):
        if not self.yes:
            return pts

        newpts = []
        for pt in pts:
            newpt = []
            for p in pt:
                newpt.append(Point(-p.x,p.y,p.z))
            newpts.append(newpt)
        return newpts

    def fmt(self, arg):
        return "%s<%s>" % (self.__class__.__name__, str(arg))

class PointMirrorY:

    def __init__(self, yes):
        self.yes = yes

    def action(self, pts):
        if not self.yes:
            return pts

        newpts = []
        for pt in pts:
            newpt = []
            for p in pt:
                newpt.append(Point(p.x,-p.y,p.z))
            newpts.append(newpt)
        return newpts

    def fmt(self, arg):
        return "%s<%s>" % (self.__class__.__name__, str(arg))

class TimeReverse:
    def __init__(self, yes):
        self.yes = yes

    def action(self, pts):
        if not self.yes:
            return pts

        newpts = []
        for pt in pts:
            newpts.append(pt)
        newpts.reverse()
        return newpts

    def fmt(self, arg):
        return "%s<%s>" % (self.__class__.__name__, str(arg))


class Rishon:

    def __init__(self, angle, chirality, zorientate=False):

        self.chirality = chirality
        self.angle = angle
        if angle is None:
            return

        self.create_points(zorientate=zorientate)
        self.ang = int(round(angle / tau * 12))
        if self.chirality == 1:
            self.title = "Rishon%dr" % self.ang
        else:
            self.title = "Rishonl%d" % self.ang

    def printout(self):
        for pt in self.points:
            print pt

    def create_points(self, pts=24, zorientate=False):
        """
        (1) start with ellipse rotation axis as Y axis (0,1,0) - [rotaxis]

        (2) rotate ellipse axis about Z by angle

        (3) Rishon starts upright (0,0,1) - [axoff] at [start]

        (4) Rishon is rotated about ellipse axis by
            1/2 angle of C-point rotation.

        (5) At same time, Rishon is rotated about C point by angle

        """

        centre = Point(0, 0, 0) # C-Point (arxiv:0910.1663)
        start = Point(0, -4, 0) # r = 4
        axoff = [Point(0, 0, 0.5), Point(0, 0, -0.5)] # rishon axis points [3]
        #axoff = [Point(0, 0, 0), Point(0, 0, 1)] # rishon axis points [3]
        #axoff = RotateX(tau/8).action_pt(axoff)
        rotaxis = [Point(0, 0, 0), Point(0, 1, 0)] # ellipse rotation axis [1]

        # rotate ellipse axis about Z by Rishon angle [2]
        rotaxis = RotateZ(self.angle).action_pt(rotaxis)
        #start = PointRotate3D(centre, Point(0,0,1), start, self.angle)

        #print "rotaxis", rotaxis
        #print "axoff", axoff
        #print "start", start

        # now create a set of 2 points, first rotating them
        # about the polarisation axis, then rotating them about C (centre)
        # where because this is mobius strip stuff, the polarisation
        # angle is 1/2 that of the rotation about C.
        self.points = []
        for i in range(pts+1): # one more, to give 0..pts not 0..pts-1
            ang = tau * float(i) / pts # we want 2 revolutions, so double
            # start [3] at centre initially
            pt = [centre + axoff[0], centre + axoff[1]]
            # rotate [3] on ellipse axis [4]
            rotang = self.chirality * ang/2 # mobius = half current angle
            r = Rotate(rotaxis, rotang) 
            pt = r.action_pt(pt)

            # whoops we're about to rotate the rotaxis below so
            # compensate for that...
            if zorientate:
                z = RotateZ(-ang)
            else:
                z = RotateZ(-ang)
            pt = z.action_pt(pt)

            # ... then move to start
            pt[0] = start + pt[0]
            pt[1] = start + pt[1]

            # rotate Rishon on zaxis about C-Point by current angle [5]
            if zorientate:
                z = RotateZ(ang - self.angle)
            else:
                z = RotateZ(ang)
            pt = z.action_pt(pt)

            self.points.append([pt[0], pt[1]])

    def _compare(self, rishon, search_for):
        if isinstance(rishon, Rishon):
            pts = rishon.points
        else:
            pts = rishon
        # 1e-6 comparison of all points
        cmp_idx = 0
        while True:
            pt = pts[search_for]
            if not self._cmp(cmp_idx, pt, False):
                return False
            cmp_idx += 1
            if cmp_idx == len(pts):
                return True
            search_for += 1
            if search_for == len(pts):
                search_for = 0

    def _compare2(self, rishon, invert=False):
        if isinstance(rishon, Rishon):
            pts = rishon.points
        else:
            pts = rishon
        # 1e-6 comparison of all points
        for i in range(len(pts)):
            for j in range(len(pts)):
                search_for = (i + j) % len(pts)
                pt = pts[search_for]
                #print "search", search_for, j
                if self._cmp2(j, pt):
                    if j == len(pts) - 1:
                        return i
                else:
                    break
        return False

    def compare(self, rishon):
        return self._compare(rishon, 0) or \
               self._compare(rishon, len(self.points)/2)

    def compare2(self, rishon):
        return self._compare2(rishon, True)

    def _cmp2(self, i, pt):
        return self.__cmp2(i, pt, True) or self.__cmp2(i, pt, False)

    def __cmp2(self, i, pt, inverse):
        pt1_1 = self.points[i][0]
        pt1_2 = self.points[i][1]
        centre1 = pt1_1.centre(pt1_2)
        centre2 = pt[0].centre(pt[1])
        if not centre1.eq(centre2):
            #print "not centre", centre1, centre2
            return False
        #print "pt1,pt2", pt1_1, pt1_2, inverse
        #print "pt[0,1]", pt[0], pt[1]
        if inverse:
            d1 = pt1_2 - pt1_1
        else:
            d1 = pt1_1 - pt1_2
        d2 = pt[0] - pt[1]
        #print "diff", d1, d2, d1.eq(d2), d1.eqinv(d2)
        return d1.eq(d2) or d1.eqinv(d2)

    def _cmp(self, i, pt, invert):
        p1l = self.points[i][0].lst() + self.points[i][1].lst()
        p2l = pt[0].lst() + pt[1].lst()
        for j in range(len(p1l)):
            if invert:
                diff = math.fabs(p1l[j] - p2l[len(p1l)-j-1])
                diff2 = math.fabs(p1l[j] - p2l[j])
                #print "compare", i, j, diff, diff2
                if diff > 1e-6 and diff2 > 1e-6:
                    #print "fail", i, j, p1l, p2l
                    return False
            else:
                diff = math.fabs(p1l[j] - p2l[j])
                #print "compare", i, j, diff
                if diff > 1e-6:
                    #print "fail", i, j, p1l, p2l
                    return False
        return True

    def __add__(self, rishon):
        newpts = []
        for i in range(len(self.points)):
            pts = []
            toadd1 = self.points[i]
            toadd2 = rishon.points[i]
            centre1 = toadd1[0].centre(toadd1[1])
            centre2 = toadd2[0].centre(toadd2[1])

            print centre1, centre2

            for j in range(2):
                t1 = toadd1[j] - centre1
                t2 = toadd2[j] - centre2
                pt = t1 + t2
                pt.scale(0.5)
                pts.append(pt + centre1)
            print pts
            newpts.append(pts)

        # no idea what the angle is (yet)
        newpt = Rishon(None, 0)
        newpt.points = newpts
        newpt.title = self.title + "+" + rishon.title

        return newpt

    def __str__(self):
        return self.title

def apply_oplist(oplist, p1, p2, _p2, applied=None):
    """
        takes a list of operations (and arguments),
        applies them in turn, compares results, records them.
    """
    if applied is None:
        applied = []
    res = []
    if len(oplist) == 0:
        return res
    oplist = copy(oplist)
    (op, args) = oplist.pop()
    p2orig = copy(_p2)
    res += apply_oplist(oplist, p1, p2, p2orig, applied) 
    for arg in args:
        print len(applied), len(args), op, arg, p1.title, _p2.title
        appliednow = copy(applied)
        oa = op(arg)
        p = oa.action(p2orig.points)
        if p1.compare(p):
            res.append((appliednow, p1, p2))
        appliednow.append((oa, arg))
        _p2 = copy(p2orig)
        _p2.points = p
        res += apply_oplist(oplist, p1, p2, _p2, appliednow)
    return res

def apply_oplist2(oplist, skipop, p1, p2, _p2, count, applied=None):
    """
        takes a list of operations (and arguments),
        apply them in sequence, apply again (except for the one you just
        did), count down until zero.
    """
    if applied is None:
        applied = []
    res = []
    if count == 0:
        return res
    for i in range(len(oplist)):
        if i == skipop:
            continue
        (op, args) = oplist[i]
        p2orig = copy(_p2)
        for arg in args:
            #print i, len(applied),len(args),count,op,arg, p1.title, _p2.title
            oa = op(arg)
            p = oa.action(p2orig.points)
            identity = False
            # check to see if this set has occurred before
            for (_oa, _arg, ap) in applied:
                if p2.compare2(ap) is not False:
                    # identity operation: we're back to start.  skip
                    #print "identity", p2.title
                    identity = True
                    break
            if identity: # no need to explore further
                continue
            appliednow = copy(applied)
            appliednow.append((oa, arg, p))
            if p1.compare2(p) is not False:
                res.append((appliednow, p1, p2))
                #print "compare ok", p1.title, p2orig.title
                #for (op, arg, p) in appliednow:
                #    print "\t", op.fmt(arg)
                continue
            _p2c = copy(p2orig)
            _p2c.points = p
            res += apply_oplist2(oplist, i, p1, p2, _p2c, count-1, appliednow)
    return res

def report_results(res):
    for (appliednow, r1, r2) in res:
        print str(r1), str(r2)
        r2p = copy(r2)
        for (op, arg, p) in appliednow:
            print "\t", op.fmt(arg)
            r2p.points = op.action(r2p.points)
        diff = r2p.compare2(r1)
        print diff
        assert diff is not False
        for i in range(3):
            plot(r1.points)
            plot(r2.points)


def frange(x, y, jump):
  while x < y:
    yield x
    x += jump

def search_equivalence(oplist, angle, chirality, timereverse, oplistlen):

    print "searching equivalence", angle, chirality, timereverse

    r = Rishon(tau * angle / 12.0, chirality)
    if timereverse:
        r.points = TimeReverse(True).action(r.points)

    for i in range(12):
        for j in [-1, 1]:
            if r.ang == i and r.chirality == j:
                continue
            r1 = Rishon(tau * i / 12.0, j)

            rs = apply_oplist2(oplist, -1, r, r1, r1, oplistlen)
            report_results( rs )

def search_equivalence_x(angle, chirality, timereverse=False, oplistlen=None):

    #tar = list(frange(tau/24, tau - tau/24, tau/24))
    oplist = []
    #oplist.append((TimeReverse, [True]))
    oplist.append((PointMirrorX, [True]))
    #oplist.append((PointMirrorZ, [True]))
    #oplist.append((PointMirrorY, [True]))
    oplist.append((RotateX, [tau/2]))
    #oplist.append((RotateX, [0, tau/2]))
    #oplist.append((RotateY, [tau/2]))
    #oplist.append((RotateZ, tar))

    if oplistlen is None:
        oplistlen = len(oplist)

    return search_equivalence(oplist, angle, chirality, timereverse, oplistlen)

def search_equivalence_y(angle, chirality, timereverse=False, oplistlen=None):

    #tar = list(frange(tau/24, tau - tau/24, tau/24))
    oplist = []
    #oplist.append((TimeReverse, [True]))
    #oplist.append((PointMirrorZ, [True]))
    oplist.append((PointMirrorY, [True]))
    oplist.append((RotateY, [tau/2]))

    if oplistlen is None:
        oplistlen = len(oplist)

    return search_equivalence(oplist, angle, chirality, timereverse, oplistlen)


def search_timereverse_y(angle, chirality, timereverse=False, oplistlen=None):

    #tar = list(frange(tau/24, tau - tau/24, tau/24))
    oplist = []
    oplist.append((PointMirrorX, [True]))
    oplist.append((RotateX, [tau/2]))

    if oplistlen is None:
        oplistlen = len(oplist)

    return search_equivalence(oplist, angle, chirality, timereverse, oplistlen)

def search_equivalence_z(angle, chirality, timereverse=False, oplistlen=None):

    #tar = list(frange(tau/24, tau - tau/24, tau/24))
    oplist = []
    oplist.append((TimeReverse, [True]))
    oplist.append((PointMirrorZ, [True]))
    oplist.append((PointMirrorX, [True]))
    #oplist.append((RotateY, [tau/2]))

    if oplistlen is None:
        oplistlen = len(oplist)

    return search_equivalence(oplist, angle, chirality, timereverse, oplistlen)


def check_symmetry_x():

    for i in range(12):
        r = Rishon(tau * i / 12.0, 1)
        r1 = Rishon(tau * i / 12.0, 1)
        r.points = PointMirrorX(True).action(r.points)
        r.points = RotateX(tau/2).action(r.points)
        diff = r1.compare2(r)
        #print "inverse diff", i, diff
        assert diff is not False, "particles not equal"

    print "no asserts, all particles equal to their 180-Rotate + MirroredX"

def check_inverse_symmetry():

    for i in range(12):
        r = Rishon(tau * i / 12.0, 1)
        r1 = Rishon(tau * ((i + 6) % 12) / 12.0, -1)
        diff = r1.compare2(r)
        #print "inverse diff", i, diff
        assert diff is not False, "particles not equal"

    print "no asserts, all particles equal to their anti-chiral anti-particle"

def check_identity_x():

    for i in range(12):
        for j in [-1, 1]:
            r = Rishon(tau * i / 12.0, j)
            r1 = Rishon(tau * i / 12.0, j)
            pts = r.points
            pts = PointMirrorX(True).action(pts)
            pts = RotateX(tau/2).action(pts)

            diff = r1.compare(pts)
            assert diff is not False, "identity XmX180 test failed"

    print "no asserts, all particles identity under X-Mirror, X-180"

def check_neutrino_y_identity():

    for i in [3,9]:
        for j in [-1, 1]:
            r = Rishon(tau * i / 12.0, j)
            r1 = Rishon(tau * (12-i) / 12.0, -j)
            pts = r.points
            diff = r1.compare(pts)
            assert diff is not False, "neutrino inverse-identity violated"
            pts = TimeReverse(True).action(pts)
            pts = PointMirrorZ(True).action(pts)
            pts = PointMirrorX(True).action(pts)

            diff = r1.compare(pts)
            assert diff is not False, \
                "neutrino identity test failed %s %s" % \
                    (str(diff), str(i))

    print "no asserts, all neutrinos identity under TR+MZ+MX"

def main():

    if False:
        for i in range(12):
            for chirality in [-1, 1]:
                r = Rishon(tau * i / 12.0, chirality, True)
                plot(r.points, r.title)

    if False:
        for j in range(2):
            for i in [3, 9]:
                r = Rishon(tau * i / 12.0, 1)
                plot(r.points)

    if False:
        for i in range(12):
            search_equivalence_x(i, 1)

    #if True:
    #    search_equivalence_x(3, -1, True, 8)

    if False:
        for i in range(12):
            search_equivalence_z(i, 1)

    if False:
        for i in range(12):
            search_timereverse_y(i, 1)

    if False:
        check_inverse_symmetry()
        check_symmetry_x()
        check_identity_x()
        check_neutrino_y_identity()

    if False:
        for i in range(12):
            search_equivalence_y(i, 1)

    if False:
        r = Rishon(tau * 0 / 12.0, 1)
        r1 = Rishon(tau * 0 / 12.0, 1)
        pts = r.points
        #pts = RotateX(tau/2).action(pts)
        ###pts = RotateY(tau/2).action(pts)
        ###pts = TimeReverse(True).action(pts)
        pts = PointMirrorY(True).action(pts)
        #pts = PointMirrorX(True).action(pts)
        ###pts = RotateX(tau/2).action(pts)

        # these two apply to neutrino anti-neutrino successfully
        plot(pts)
        #pts = TimeReverse(True).action(pts)
        #pts = PointMirrorZ(True).action(pts)
        #pts = PointMirrorX(True).action(pts)
        #plot(pts)
        #pts = RotateX(tau/2).action(pts)
        #pts = PointMirrorZ(True).action(pts)
        #pts = TimeReverse(True).action(pts)
        pts = RotateY(tau/2).action(pts)
        #plot(pts)
        for x in zip(r1.points, pts):
            print x
        plot(r1.points)
        plot(pts)

        print r1.compare2(pts)

    if True:
        r = Rishon(tau * 1 / 12.0, 1, True)
        r1 = Rishon(tau * 8 / 12.0, 1, True)

        r.points = RotateZ(3*tau/12).action(r.points)
        r1.points = RotateZ(-4*tau/12).action(r1.points)
        while True:
            plot(r.points)
            plot(r1.points)
            plot(r.points + r1.points)

    if False:
        r = Rishon(tau * 2 / 12.0, 1, True)
        r1 = Rishon(tau * 9 / 12.0, 1, True)
        #r1.points = RotateZ(6*tau/12).action(r1.points)
        #r1.points = RotateX(6*tau/12).action(r1.points)
        #r.points = RotateZ(-1*tau/12).action(r.points)
        #r1.points = RotateZ(-8*tau/12).action(r1.points)
        radd = r + r1
    
        #radd.points = RotateZ(-1*tau/12).action(radd.points)
        #radd.points = PointMirrorZ(True).action(radd.points)
        for i in range(len(radd.points)):
            print "%-20s" % r.title, r.points[i]
            print "%-20s" % radd.title, radd.points[i]
            radd._cmp2(i, r.points[i])

        print r.compare2(radd)

        print radd
        print radd.points[0]
        print r.points[0]
        print r1.points[0]
        print radd.points[1]
        print r.points[1]
        print r1.points[1]
        while False:
            plot(r.points)
            plot(r1.points)
            plot(radd.points)

        for i in range(12):
            for j in [-1, 1]:
                r2 = Rishon(tau * i / 12.0, j, True)
                #r2.points = TimeReverse(True).action(r2.points)
                #r2.points = PointMirrorZ(True).action(r2.points)
                diff = r2.compare2(radd)
                print diff
                if diff is not False:
                    print "found", i, j, radd.title, r2.title, diff

    while False:
        for i in [0, 6, 3, 9]:
        #for i in [1, 7, 4, 10]:
            r = Rishon(tau * i / 12.0)
            plot(r.points)

# when executed, just run main():
if __name__ == '__main__':
    main()

